"""
ratus - A simple expression language.

ratus is a simple expression language intended to be used to easily and safely
extend applications in a controllable way. It provides the following features:

- Callable functions

  - These can be injected when evaluating the expression but they cannot be
    defined within the expression language
  - ``if`` is provided by default but can be overridden if you really want to

- Simple math operations, more advanced functions can be implemented as
  functions

  - Addition (``+``)
  - Subtractions (``-``)
  - Multiplication (``*``)
  - Division (``/``)

- Comparison operators

  - Equal (``=``)
  - Not equal (``!=``)
  - Greater than (``>``)
  - Greater than or equal (``>=``)
  - Less than (``<``)
  - Less than or equal (``<=``)

- Literals

  - String (double and single quotes)
  - Integer (positive and negative)
  - Float (positive and negative)
"""
from typing import Any, Callable, Dict, Optional

from ratus.execer import Executor
from ratus.parse import Parser
from ratus.token import Tokeniser

__version__ = "0.0.1"

__all__ = ["Evaluator", "token", "parse", "execer"]


class Evaluator:
    "Expression evaluator."

    def __init__(
        self, injected_functions: Optional[Dict[str, Callable[..., Any]]] = None
    ) -> None:
        self.tokeniser = Tokeniser()
        self.parser = Parser()
        self.executor = Executor(injected_functions)

    def evaluate(self, source: str) -> Any:
        """Evaluate an input as a ratus expression."""
        tokens = self.tokeniser.tokenise(source)

        expression = self.parser.parse(tokens)

        return self.executor.execute(expression)



from abc import ABC
from dataclasses import dataclass
from enum import Enum
from typing import Any, List, Tuple, cast

from ratus.token import Token, TokenLiteral, TokenType


class ParserError(Exception):
    """Exception raised when there is an error parsing."""


class Expression(ABC):
    """Base representation of an expression."""


class Literal(Expression, ABC):
    """Literal value."""

    value: Any


@dataclass
class Integer(Literal):
    """Integer literal."""

    value: int


@dataclass
class Float(Literal):
    """Float literal."""

    value: float


@dataclass
class String(Literal):
    """String literal."""

    value: str


@dataclass
class Function(Expression):
    """
    Function call.

    Functions have a `name` which is used to look them up in the executor and a
    list of arguments that the function should be called with.
    """

    name: str
    args: List[Expression]


class BinaryOpType(Enum):
    """Binary operation type."""

    ADDITION = "+"
    SUBTRACTION = "-"
    MULTIPLICATION = "*"
    DIVISION = "/"
    EQUAL = "="
    NOT_EQUAL = "!="
    GREATER = ">"
    GREATER_EQUAL = ">="
    LESS = "<"
    LESS_EQUAL = "<="
    AND = "and"
    OR = "or"


@dataclass
class BinaryOp(Expression, ABC):
    """
    Binary operation.

    Binary operations have a type that maps them to the operation that is being
    performed and left and right operands upon which the operation is performed.
    """

    op_type: BinaryOpType
    left: Expression
    right: Expression


class UnaryOpType(Enum):
    """Unary operation type."""

    NOT = "!"
    NEGATIVE = "-"


@dataclass
class UnaryOp(Expression, ABC):
    """
    Unary operation.

    Unary operations have a type and a single operand upon which the action is
    performed.
    """

    op_type: UnaryOpType
    operand: Expression


def _parse(tokens: List[Token]) -> Expression:
    expression, _ = _parse_expression(tokens)
    return expression


def _parse_expression(tokens: List[Token]) -> Tuple[Expression, List[Token]]:
    if len(tokens) == 0:
        raise ParserError("Expression cannot be empty")
    if tokens[0].token_type is TokenType.STRING and isinstance(tokens[0], TokenLiteral):
        return String(tokens[0].literal), tokens[1:]
    if tokens[0].token_type is TokenType.IDENT and isinstance(tokens[0], TokenLiteral):
        return _parse_function(tokens)
    expr, rest = _parse_term(tokens)
    while len(rest) > 0:
        operator, *rest = rest
        if operator.token_type not in (
            TokenType.PLUS,
            TokenType.MINUS,
            TokenType.GREATER,
            TokenType.GREATER_EQUAL,
            TokenType.LESS,
            TokenType.LESS_EQUAL,
            TokenType.EQUAL,
            TokenType.BANG_EQUAL,
            TokenType.AND,
            TokenType.OR,
        ):
            raise ParserError(
                f"Unexpected token after term {expr}. Expected operator '+', "
                "'-', '>', '>='. '<'. '<=', '=', '!=', 'and', 'or'."
            )
        right_term, rest = _parse_term(rest)
        operator_type_mapping = {
            TokenType.PLUS: BinaryOpType.ADDITION,
            TokenType.MINUS: BinaryOpType.SUBTRACTION,
            TokenType.GREATER: BinaryOpType.GREATER,
            TokenType.GREATER_EQUAL: BinaryOpType.GREATER_EQUAL,
            TokenType.LESS: BinaryOpType.LESS,
            TokenType.LESS_EQUAL: BinaryOpType.LESS_EQUAL,
            TokenType.EQUAL: BinaryOpType.EQUAL,
            TokenType.BANG_EQUAL: BinaryOpType.NOT_EQUAL,
            TokenType.AND: BinaryOpType.AND,
            TokenType.OR: BinaryOpType.OR,
        }
        operator_type = operator_type_mapping[operator.token_type]
        expr = BinaryOp(operator_type, expr, right_term)
    return expr, []


def _parse_function(tokens: List[Token]) -> Tuple[Expression, List[Token]]:
    if len(tokens) < 3:
        raise ParserError(f"Tokens {tokens} do not form a valid function call")
    if tokens[1].token_type is not TokenType.LEFT_PAREN:
        raise ParserError(
            f"Expected left paren ('(') following call to function '{tokens[0].lexeme}'. Found '{tokens[1].lexeme}'"
        )
    nesting = 0
    end_of_function_call_idx = 0
    for i, token in enumerate(tokens):
        if token.token_type is TokenType.LEFT_PAREN:
            nesting += 1
        elif token.token_type is TokenType.RIGHT_PAREN:
            nesting -= 1
            if nesting == 0:
                end_of_function_call_idx = i
                break
    else:
        raise ParserError(
            f"Unbalanced parentheses in call to function '{tokens[0].lexeme}'"
        )
    func_token: TokenLiteral
    func_token = cast(TokenLiteral, tokens[0])
    name = func_token.literal
    if len(tokens) == 3:
        return Function(name, args=[]), tokens[end_of_function_call_idx + 1 :]
    args_tokens, rest = _split_function_args(tokens[2:])
    args = [_parse(arg_tokens) for arg_tokens in args_tokens]
    return Function(name, args=args), rest


def _parse_term(tokens: List[Token]) -> Tuple[Expression, List[Token]]:
    term, rest = _parse_factor(tokens)
    while len(rest) > 0:
        operator, *rest = rest
        if operator.token_type not in (TokenType.STAR, TokenType.SLASH):
            return term, [operator] + rest

        # Based on the condition we know the operator token is STAR or SLASH
        right_factor, rest = _parse_factor(rest)
        operator_type_mapping = {
            TokenType.STAR: BinaryOpType.MULTIPLICATION,
            TokenType.SLASH: BinaryOpType.DIVISION,
        }
        operator_type = operator_type_mapping[operator.token_type]
        term = BinaryOp(operator_type, term, right_factor)

    return term, rest


def _parse_factor(tokens: List[Token]) -> Tuple[Expression, List[Token]]:
    if tokens[0].token_type is TokenType.LEFT_PAREN:
        expr_end_index = 0
        nesting = 0
        for i, token in enumerate(tokens):
            if token.token_type is TokenType.LEFT_PAREN:
                nesting += 1
            elif token.token_type is TokenType.RIGHT_PAREN:
                nesting -= 1
                if nesting == 0:
                    expr_end_index = i
                    break
        expr_tokens = tokens[1:expr_end_index]
        expr = _parse(expr_tokens)
        return expr, tokens[expr_end_index + 1 :]
    return _parse_number(tokens)


def _parse_number(
    tokens: List[Token], bang_allowed=True, minus_allowed=True
) -> Tuple[Expression, List[Token]]:
    if len(tokens) == 0:
        raise ParserError(f"Expected int or float token but none were found")
    if tokens[0].token_type is TokenType.INT and isinstance(tokens[0], TokenLiteral):
        return Integer(tokens[0].literal), tokens[1:]
    if tokens[0].token_type is TokenType.FLOAT and isinstance(tokens[0], TokenLiteral):
        return Float(tokens[0].literal), tokens[1:]
    if tokens[0].token_type is TokenType.MINUS and minus_allowed:
        operand, rest = _parse_number(tokens[1:], bang_allowed=False)
        return UnaryOp(UnaryOpType.NEGATIVE, operand), rest
    if tokens[0].token_type is TokenType.BANG and bang_allowed:
        operand, rest = _parse_number(tokens[1:], minus_allowed=False)
        return UnaryOp(UnaryOpType.NOT, operand), rest
    raise ParserError(f"Unexpected token {tokens[0]}. Expected an int or float")


def _split_function_args(tokens: List[Token]) -> Tuple[List[List[Token]], List[Token]]:
    nesting = 1
    arg: List[Token] = []
    function_args: List[List[Token]] = []
    while len(tokens) > 0:
        token = tokens.pop(0)
        if token.token_type is TokenType.LEFT_PAREN:
            nesting += 1
        elif token.token_type is TokenType.RIGHT_PAREN:
            nesting -= 1
            if nesting == 0:
                function_args.append(arg)
                return function_args, tokens

        if nesting == 1 and token.token_type is TokenType.COMMA:
            function_args.append(arg)
            arg = []
        else:
            arg.append(token)
    raise ParserError("Function call does not have closing paren (')')")


class Parser:
    def parse(self, tokens: List[Token]):
        return _parse(tokens)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['ratus']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'ratus',
    'version': '0.0.1',
    'description': 'Simple expression language.',
    'long_description': None,
    'author': 'Nick Spain',
    'author_email': 'nicholas.spain96@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import operator
from typing import Any, Callable, Dict, Optional

from ratus.parse import BinaryOp, Expression, Function, Literal, UnaryOp


class ExecutorError(Exception):
    """Exception raised if there is an error executing an expression."""


class Executor:
    """Executor of expressions."""

    def __init__(
        self,
        functions: Optional[Dict[str, Callable[..., Any]]] = None,
        binary_ops: Optional[Dict[str, Callable[[Any, Any], Any]]] = None,
        unary_ops: Optional[Dict[str, Callable[[Any], Any]]] = None,
    ) -> None:
        """
        Instantiate an Executor object.

        This constructor provides three arguments to control behavior.

        `functions` allows us to extend the callable functions in expression. By default
        only `if` is provided but more can be added and the default definition of `if`
        (`lambda c, s, f: s if c else f`) can be overridden simple by having an "if" key
        in the `functions` dictionary.

        `binary_ops` allows us to extend the binary operations available. It is a
        dictionary mapping variants of `ratus.parse.BinaryOpTypes` to a function with
        two parameters and a single output.

        `unary_ops` allows us to extend the unary operations available. It is a
        dictionary mapping variants of `ratus.parse.UnaryOpTypes` to a function with one
        parameter and one outputs.
        """
        self.binary_ops: Dict[str, Callable[[Any, Any], Any]] = {
            "+": operator.add,
            "-": operator.sub,
            "*": operator.mul,
            "/": operator.truediv,
            ">": operator.gt,
            ">=": operator.ge,
            "<": operator.lt,
            "<=": operator.le,
            "and": operator.and_,
            "or": operator.or_,
            "=": operator.eq,
            "!=": operator.ne,
        }
        if binary_ops is not None:
            self.binary_ops.update(binary_ops)

        self.unary_ops: Dict[str, Callable[[Any], Any]] = {
            "!": operator.not_,
            "-": operator.neg,
        }
        if unary_ops is not None:
            self.unary_ops.update(unary_ops)

        self.functions = {"if": lambda c, s, f: s if c else f}
        if functions is not None:
            self.functions.update(functions)

    def execute(self, expression: Expression) -> Any:
        """Execute an expression and return the result."""
        if isinstance(expression, Literal):
            return expression.value
        if isinstance(expression, BinaryOp):
            left = self.execute(expression.left)
            right = self.execute(expression.right)
            binary_op = self.binary_ops[expression.op_type.value]
            return binary_op(left, right)
        if isinstance(expression, UnaryOp):
            operand = self.execute(expression.operand)
            unary_op = self.unary_ops[expression.op_type.value]
            return unary_op(operand)
        if isinstance(expression, Function):
            function = self.functions.get(expression.name)
            if function is None:
                raise ExecutorError(f"Function '{expression.name}' is not defined")
            args = [self.execute(arg) for arg in expression.args]
            return function(*args)



from dataclasses import dataclass
from enum import Enum
from typing import Any, List


class TokeniserError(Exception):
    """Error raised in the tokeniser."""


class TokenType(Enum):
    """Token type."""

    LEFT_PAREN = "("
    RIGHT_PAREN = ")"
    COMMA = ","
    PLUS = "+"
    MINUS = "-"
    SLASH = "/"
    STAR = "*"

    BANG = "!"
    EQUAL = "="
    BANG_EQUAL = "!="
    GREATER = ">"
    GREATER_EQUAL = ">="
    LESS = "<"
    LESS_EQUAL = "<="

    INT = "int"
    FLOAT = "float"
    STRING = "string"
    IDENT = "ident"

    AND = "and"
    OR = "or"

    EOF = "eof"


@dataclass
class Token:
    """Representation of a token."""

    token_type: TokenType
    lexeme: str


@dataclass
class TokenLiteral(Token):
    """Representation of a literal."""

    literal: Any


class Tokeniser:
    def __init__(self):
        self.start: int = 0
        self.current: int = 0
        self.source: str = ""
        self.tokens: List[Token] = []

    def tokenise(self, source: str) -> List[Token]:
        """Tokenise an input in a list of tokens."""
        self.source = source
        self.current = 0
        self.start = 0
        while self.current < len(self.source):
            self.start = self.current
            self.scan_token()
        return self.tokens

    def scan_token(self):
        c = self.source[self.current]
        self.current += 1
        if c.strip() == "":
            # Skip whitespace
            return
        if c in ("(", ")", ",", "+", "-", "*", "=", "/"):
            # Match characters that are unambiguously only single characters
            self.add_token(TokenType(c))
        elif c == "!":
            if self.source[self.current] == "=":
                self.add_token(TokenType.BANG_EQUAL)
                self.current += 1
            else:
                self.add_token(TokenType.BANG)
        elif c == "<":
            if self.source[self.current] == "=":
                self.add_token(TokenType.LESS_EQUAL)
                self.current += 1
            else:
                self.add_token(TokenType.LESS)
        elif c == ">":
            if self.source[self.current] == "=":
                self.add_token(TokenType.GREATER_EQUAL)
                self.current += 1
            else:
                self.add_token(TokenType.GREATER)
        elif c in ("'", '"'):
            self.string()
        elif c.isdigit():
            self.numeric()
        elif c.isalpha():
            self.identifier()
        else:
            raise TokeniserError(f"Unexpected character: '{c}'")

    def add_token(self, token_type: TokenType):
        lexeme = self.source[self.start : self.current]
        token = Token(token_type, lexeme)
        self.tokens.append(token)

    def string(self):
        while self.current < len(self.source) and self.source[self.current] not in (
            "'",
            '"',
        ):
            self.current += 1
        if self.current >= len(self.source):
            raise TokeniserError("Unterminated string")
        self.current += 1  # Consume closing quote
        lexeme = self.source[self.start : self.current]
        string = self.source[self.start + 1 : self.current - 1]
        token = TokenLiteral(TokenType.STRING, lexeme, string)
        self.tokens.append(token)

    def numeric(self):
        while self.current < len(self.source) and self.source[self.current].isdigit():
            self.current += 1

        # Invalid to finish expression with "."
        if self.current == len(self.source) - 1 and self.source[self.current] == ".":
            raise TokeniserError("Expression cannot finish with '.'")

        if self.current < len(self.source) and self.source[self.current] == ".":
            # Consume the "." so we can start consuming digits again
            self.current += 1
            if not self.source[self.current].isdigit():
                raise TokeniserError(
                    f"Expected digit after '.', found '{self.source[self.current]}'"
                )

            # Match a float
            while (
                self.current < len(self.source) and self.source[self.current].isdigit()
            ):
                self.current += 1
            float_ = self.source[self.start : self.current]
            token = TokenLiteral(TokenType.FLOAT, float_, float(float_))
            self.tokens.append(token)
        else:
            int_ = self.source[self.start : self.current]
            token = TokenLiteral(TokenType.INT, int_, int(int_))
            self.tokens.append(token)

    def identifier(self):
        #  Identifiers can be made up of letters, numbers and '_'
        while self.current < len(self.source) and (
            self.source[self.current].isalpha()
            or self.source[self.current].isdigit()
            or self.source[self.current] == "_"
        ):
            self.current += 1
        ident = self.source[self.start : self.current]
        token = TokenLiteral(TokenType.IDENT, ident, ident)
        self.tokens.append(token)
